@@ -23,6 +23,7 @@
 
 #include "byte_order.h"
 #include "rhash_timing.h"
+#include "rhash_torrent.h"
 
 #ifdef USE_RHASH_DLL
 # define RHASH_API __declspec(dllimport)
@@ -31,9 +32,11 @@
 #include "test_hashes.h"
 
 /*=========================================================================*
- *                             Data for tests                              *
+ *                              Test vectors                               *
  *=========================================================================*/
-const char* crc32_tests[] = { /* verified with cksfv */
+
+ /* verified by cksfv */
+ const char* crc32_tests[] = {
 	"", "00000000",
 	"a", "E8B7BE43",
 	"abc", "352441C2",
@@ -45,6 +48,18 @@ const char* crc32_tests[] = { /* verified with cksfv */
 	0
 };
 
+ const char* crc32c_tests[] = {
+	"", "00000000",
+	"a", "C1D04330",
+	"abc", "364B3FB7",
+	"message digest", "02BD79D0",
+	"abcdefghijklmnopqrstuvwxyz", "9EE6EF25",
+	"The quick brown fox jumps over the lazy dog", "22620404",
+	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789", "A245D57D",
+	"12345678901234567890123456789012345678901234567890123456789012345678901234567890", "477A6781",
+	0
+};
+
 const char* md4_tests[] = {
 	"", "31D6CFE0D16AE931B73C59D7E0C089C0",
 	"a", "BDE52CB31DE33E46245E05FBDBD6FB24",
@@ -57,7 +72,7 @@ const char* md4_tests[] = {
 	0
 };
 
-/* for short messages ed2k hash values are equal to md4 */
+/* for short messages ed2k test vectors coincide with md4 */
 #define ed2k_tests md4_tests
 
 /* test vectors from spec */
@@ -163,12 +178,12 @@ const char* ripemd_tests[] = {
 };
 
 /*
-* Two important test-cases (some sites calculate them incorrectly):
+* Two important test-cases (some libraries calculate them incorrectly):
 * GOST( <100000 characters of 'a'> ) = 5C00CCC2734CDD3332D3D4749576E3C1A7DBAF0E7EA74E9FA602413C90A129FA
 * GOST( <128 characters of 'U'> ) = 53A3A3ED25180CEF0C1D85A074273E551C25660A87062A52D926A9E8FE5733A4
 */
 
-/* test vectors from internet, verified by openssl and some other progs */
+/* test vectors from internet, verified by OpenSSL and some other programs */
 const char* gost_tests[] = {
 	"", "CE85B99CC46752FFFEE35CAB9A7B0278ABB4C2D2055CFF685AF4912C49490F8D",
 	"a", "D42C539E367C66E9C88A801F6649349C21871B4344C6A573F849FDCE62F314DD",
@@ -293,6 +308,52 @@ const char* sha512_tests[] = {
 
 /* SHA3 test vectors were verified by the reference implementation of Keccak */
 const char* sha3_224_tests[] = {
+	"", "6B4E03423667DBB73B6E15454F0EB1ABD4597F9A1B078E3F5B5A6BC7",
+	"a", "9E86FF69557CA95F405F081269685B38E3A819B309EE942F482B6A8B",
+	"abc", "E642824C3F8CF24AD09234EE7D3C766FC9A3A5168D0C94AD73B46FDF",
+	"message digest", "18768BB4C48EB7FC88E5DDB17EFCF2964ABD7798A39D86A4B4A1E4C8",
+	"abcdefghijklmnopqrstuvwxyz", "5CDECA81E123F87CAD96B9CBA999F16F6D41549608D4E0F4681B8239",
+	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789", "A67C289B8250A6F437A20137985D605589A8C163D45261B15419556E",
+	"12345678901234567890123456789012345678901234567890123456789012345678901234567890", "0526898E185869F91B3E2A76DD72A15DC6940A67C8164A044CD25CC8",
+	"The quick brown fox jumps over the lazy dog", "D15DADCEAA4D5D7BB3B48F446421D542E08AD8887305E28D58335795",
+	0
+};
+const char* sha3_256_tests[] = {
+	"", "A7FFC6F8BF1ED76651C14756A061D662F580FF4DE43B49FA82D80A4B80F8434A",
+	"a", "80084BF2FBA02475726FEB2CAB2D8215EAB14BC6BDD8BFB2C8151257032ECD8B",
+	"abc", "3A985DA74FE225B2045C172D6BD390BD855F086E3E9D525B46BFE24511431532",
+	"message digest", "EDCDB2069366E75243860C18C3A11465ECA34BCE6143D30C8665CEFCFD32BFFD",
+	"abcdefghijklmnopqrstuvwxyz", "7CAB2DC765E21B241DBC1C255CE620B29F527C6D5E7F5F843E56288F0D707521",
+	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789", "A79D6A9DA47F04A3B9A9323EC9991F2105D4C78A7BC7BEEB103855A7A11DFB9F",
+	"12345678901234567890123456789012345678901234567890123456789012345678901234567890", "293E5CE4CE54EE71990AB06E511B7CCD62722B1BEB414F5FF65C8274E0F5BE1D",
+	"The quick brown fox jumps over the lazy dog", "69070DDA01975C8C120C3AADA1B282394E7F032FA9CF32F4CB2259A0897DFC04",
+	0
+};
+const char* sha3_384_tests[] = {
+	"", "0C63A75B845E4F7D01107D852E4C2485C51A50AAAA94FC61995E71BBEE983A2AC3713831264ADB47FB6BD1E058D5F004",
+	"a", "1815F774F320491B48569EFEC794D249EEB59AAE46D22BF77DAFE25C5EDC28D7EA44F93EE1234AA88F61C91912A4CCD9",
+	"abc", "EC01498288516FC926459F58E2C6AD8DF9B473CB0FC08C2596DA7CF0E49BE4B298D88CEA927AC7F539F1EDF228376D25",
+	"message digest", "D9519709F44AF73E2C8E291109A979DE3D61DC02BF69DEF7FBFFDFFFE662751513F19AD57E17D4B93BA1E484FC1980D5",
+	"abcdefghijklmnopqrstuvwxyz", "FED399D2217AAF4C717AD0C5102C15589E1C990CC2B9A5029056A7F7485888D6AB65DB2370077A5CADB53FC9280D278F",
+	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789", "D5B972302F5080D0830E0DE7B6B2CF383665A008F4C4F386A61112652C742D20CB45AA51BD4F542FC733E2719E999291",
+	"12345678901234567890123456789012345678901234567890123456789012345678901234567890", "3C213A17F514638ACB3BF17F109F3E24C16F9F14F085B52A2F2B81ADC0DB83DF1A58DB2CE013191B8BA72D8FAE7E2A5E",
+	"The quick brown fox jumps over the lazy dog", "7063465E08A93BCE31CD89D2E3CA8F602498696E253592ED26F07BF7E703CF328581E1471A7BA7AB119B1A9EBDF8BE41",
+	0
+};
+const char* sha3_512_tests[] = {
+	"", "A69F73CCA23A9AC5C8B567DC185A756E97C982164FE25859E0D1DCC1475C80A615B2123AF1F5F94C11E3E9402C3AC558F500199D95B6D3E301758586281DCD26",
+	"a", "697F2D856172CB8309D6B8B97DAC4DE344B549D4DEE61EDFB4962D8698B7FA803F4F93FF24393586E28B5B957AC3D1D369420CE53332712F997BD336D09AB02A",
+	"abc", "B751850B1A57168A5693CD924B6B096E08F621827444F70D884F5D0240D2712E10E116E9192AF3C91A7EC57647E3934057340B4CF408D5A56592F8274EEC53F0",
+	"message digest", "3444E155881FA15511F57726C7D7CFE80302A7433067B29D59A71415CA9DD141AC892D310BC4D78128C98FDA839D18D7F0556F2FE7ACB3C0CDA4BFF3A25F5F59",
+	"abcdefghijklmnopqrstuvwxyz", "AF328D17FA28753A3C9F5CB72E376B90440B96F0289E5703B729324A975AB384EDA565FC92AADED143669900D761861687ACDC0A5FFA358BD0571AAAD80ACA68",
+	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789", "D1DB17B4745B255E5EB159F66593CC9C143850979FC7A3951796ABA80165AAB536B46174CE19E3F707F0E5C6487F5F03084BC0EC9461691EF20113E42AD28163",
+	"12345678901234567890123456789012345678901234567890123456789012345678901234567890", "9524B9A5536B91069526B4F6196B7E9475B4DA69E01F0C855797F224CD7335DDB286FD99B9B32FFE33B59AD424CC1744F6EB59137F5FB8601932E8A8AF0AE930",
+	"The quick brown fox jumps over the lazy dog", "01DEDD5DE4EF14642445BA5F5B97C15E47B9AD931326E4B0727CD94CEFC44FFF23F07BF543139939B49128CAF436DC1BDEE54FCB24023A08D9403F9B4BF0D450",
+	0
+};
+
+#ifdef USE_KECCAK
+const char* keccak_224_tests[] = {
 	"", "F71837502BA8E10837BDD8D365ADB85591895602FC552B48B7390ABD",
 	"a", "7CF87D912EE7088D30EC23F8E7100D9319BFF090618B439D3FE91308",
 	"abc", "C30411768506EBE1C2871B1EE2E87D38DF342317300A9B97A95EC6A8",
@@ -304,7 +365,7 @@ const char* sha3_224_tests[] = {
 	0
 };
 
-const char* sha3_256_tests[] = {
+const char* keccak_256_tests[] = {
 	"", "C5D2460186F7233C927E7DB2DCC703C0E500B653CA82273B7BFAD8045D85A470",
 	"a", "3AC225168DF54212A25C1C01FD35BEBFEA408FDAC2E31DDD6F80A4BBF9A5F1CB",
 	"abc", "4E03657AEA45A94FC7D47BA826C8D667C0D1E6E33A64A036EC44F58FA12D6C45",
@@ -316,7 +377,7 @@ const char* sha3_256_tests[] = {
 	0
 };
 
-const char* sha3_384_tests[] = {
+const char* keccak_384_tests[] = {
 	"", "2C23146A63A29ACF99E73B88F8C24EAA7DC60AA771780CCC006AFBFA8FE2479B2DD2B21362337441AC12B515911957FF",
 	"a", "85E964C0843A7EE32E6B5889D50E130E6485CFFC826A30167D1DC2B3A0CC79CBA303501A1EEABA39915F13BAAB5ABACF",
 	"abc", "F7DF1165F033337BE098E7D288AD6A2F74409D7A60B49C36642218DE161B1F99F8C681E4AFAF31A34DB29FB763E3C28E",
@@ -328,7 +389,7 @@ const char* sha3_384_tests[] = {
 	0
 };
 
-const char* sha3_512_tests[] = {
+const char* keccak_512_tests[] = {
 	"", "0EAB42DE4C3CEB9235FC91ACFFE746B29C29A8C366B7C60E4E67C466F36A4304C00FA9CAF9D87976BA469BCBE06713B435F091EF2769FB160CDAB33D3670680E",
 	"a", "9C46DBEC5D03F74352CC4A4DA354B4E9796887EEB66AC292617692E765DBE400352559B16229F97B27614B51DBFBBB14613F2C10350435A8FEAF53F73BA01C7C",
 	"abc", "18587DC2EA106B9A1563E32B3312421CA164C7F1F07BC922A9C83D77CEA3A1E5D0C69910739025372DC14AC9642629379540C17E2A65B19D77AA511A9D00BB96",
@@ -339,6 +400,7 @@ const char* sha3_512_tests[] = {
 	"The quick brown fox jumps over the lazy dog", "D135BB84D0439DBAC432247EE573A23EA7D3C9DEB2A968EB31D47C4FB45F1EF4422D6C531B5B9BD6F449EBCC449EA94D0A8F05F62130FDA612DA53C79659F609",
 	0
 };
+#endif /* USE_KECCAK */
 
 /* verified by eBASH SUPERCOP implementation */
 const char* edonr256_tests[] = {
@@ -368,48 +430,55 @@ const char* edonr512_tests[] = {
 	0
 };
 
-#define USE_BTIH_WITH_TEST_FILENAME
-
-const char* btih_tests[] = {
-#ifdef USE_BTIH_WITH_TEST_FILENAME
-	/* BTIH calculated with filename = "test.txt", verified using uTorrent */
-	"", "B9E3C388E3D6B5F63EA5CE61F79C23E2C50EF95B",
-	"a", "C277260F7EC3DAB6D15F9B38796578029ED9B32B",
-	"abc", "D8C4B957D2EEB55D39B2FE68AB924EB007355050",
-	"message digest", "8FE65F36D46DEAB1A347CE005A025F302D35D408",
-	"abcdefghijklmnopqrstuvwxyz", "2932DC7AA83123F822E9431F8F10F2C740D0EDF0",
-	"The quick brown fox jumps over the lazy dog", "C39D1F8950FDB39C2366E9A6CD7814AC3686CA55",
-	"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq", "4405635D7611838250EEF66AA8F03980C42AD2CB",
-	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789", "A6583B70D9634170940AEEDFE2AD37DADC99B5D4",
-	"12345678901234567890123456789012345678901234567890123456789012345678901234567890", "545472D099CE50D075A318429C48BE54DDA02731",
-#else
-	/* BTIH calculated without a filename, can't be verified by torrent tools */
-	"", "A66A2FA70401A9DCA25C7C39937921D377A24C1A",
-	"a", "FD408E9D024B58A57AA1313EFF14005FF8B2C5D1",
-	"abc", "83FFAAFFA483B91FBC284D1930F2DEE886515AFB",
-	"message digest", "8208E521BBECEA61711F787824FB301ED3E2C16C",
-	"abcdefghijklmnopqrstuvwxyz", "47D855AE785FEE80A210DF1512B72AE96A50C99E",
-	"The quick brown fox jumps over the lazy dog", "987147B78B586249263F89E8E7C4E636225BC1F7",
-	"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq", "5E4886F153F905528055DEB7C5181F459064D9F4",
-	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789", "867A6364D160679F8F7C41C126EAFCB9E8759CE6",
-	"12345678901234567890123456789012345678901234567890123456789012345678901234567890", "1BFF44EA31BDF301B9E02689815295C4BB25635F",
-#endif
+/* BTIH calculated with filename = "test.txt", verified using uTorrent */
+const char* btih_with_filename_tests[] = {
+	"", "042C8E2D2780B0AFAE6599A02914D6C3F1515B12",
+	"a", "7527A903193C87093C05DE0F0F81126A4B98EE1A",
+	"abc", "CBF4F6D5CCDE0E6DD6BC8F013AA7F920900C11A2",
+	"message digest", "FFFCE897C2D5FB8ED4B6AD773CC0FFA071AEC393",
+	"abcdefghijklmnopqrstuvwxyz", "606A31B06B17547C226C9EA8EE00EEBA6E0E5BFC",
+	"The quick brown fox jumps over the lazy dog", "1EED1B4C56186456E3DE420FE69A67F53CA6A52A",
+	"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq", "6BD6F0B19FA3F54CE0311BF6D2D6D3955B1BD20C",
+	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789", "95880C5A8EB06C1AC28BA6A531505E0F2BCD77AE",
+	"12345678901234567890123456789012345678901234567890123456789012345678901234567890", "674E17AA21981A33892119E601DF3E2E689C4E62",
 	0
 };
 
-typedef struct known_strings_t {
-	unsigned hash_id;
-	const char** tests;
-} known_strings_t;
+/* BTIH calculated without a filename, can't be verified by torrent tools */
+const char* btih_without_a_filename_tests[] = {
+	"", "A4A6678B3A933D1D9A182CC38E73124F7672C7EB",
+	"a", "827CD89846FC132E2E67E29C2784C65443BB4DC1",
+	"abc", "88713704608141F4E86F58C86247CA1E3D91D864",
+	"message digest", "C654901E82A8FC13C343271520FAF52EBEEF2183",
+	"abcdefghijklmnopqrstuvwxyz", "4AE72ED186D196D8F117E449D34E216B0941FF61",
+	"The quick brown fox jumps over the lazy dog", "A9328020229C163BBF07181C8DF37CE84FC66589",
+	"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq", "DB9A7E577A346FF058D78576102F2EB9DC849018",
+	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789", "7F8FEED959E89BB097B5F741C0342B6DF0A7864D",
+	"12345678901234567890123456789012345678901234567890123456789012345678901234567890", "92D6F01285FDA7D43E7D09A5AFF01C383CEEE610",
+	0
+};
 
-known_strings_t known_strings[] = {
+/** Set of test vectors for one hash function */
+struct test_vectors_t {
+	/** Hash function id */
+	unsigned hash_id;
+	/** Array of pairs (message, message_digest) */
+	const char** tests;
+};
+
+/**
+ * Array of test vectors for short messages
+ */
+struct test_vectors_t short_test_vectors[] = {
 	{ RHASH_CRC32, crc32_tests },
+	{ RHASH_CRC32C, crc32c_tests },
 	{ RHASH_MD4, md4_tests },
 	{ RHASH_MD5, md5_tests },
 	{ RHASH_SHA1, sha1_tests },
 	{ RHASH_TIGER, tiger_hashes },
 	{ RHASH_TTH, tth_tests },
-	{ RHASH_BTIH, btih_tests },
+	{ RHASH_BTIH, btih_with_filename_tests },
+	{ RHASH_BTIH, btih_without_a_filename_tests },
 	{ RHASH_ED2K, ed2k_tests },
 	{ RHASH_AICH, aich_tests },
 	{ RHASH_WHIRLPOOL, whirlpool_tests },
@@ -433,31 +502,24 @@ known_strings_t known_strings[] = {
 };
 
 /*=========================================================================*
- *                    Helper functions to hash messages                    *
+ *               Functions for calculating a message digest                *
  *=========================================================================*/
 
-static int g_errors = 0;  /* total number of errors occured */
-
-#ifdef UNDER_CE /* if Windows CE */
-static char *g_msg = NULL; /* string buffer to store errors */
-#endif
+/**
+ * The total number of errors
+ */
+static int g_errors = 0;
 
+/**
+ * Print a formatted message to the message log.
+ * @param format the format of the message
+ */
 static void log_message(char* format, ...)
 {
 	va_list vl;
-#ifndef UNDER_CE /* not Windows CE */
 	va_start(vl, format);
 	vprintf(format, vl);
 	fflush(stdout);
-#else
-	char str[300];
-	int add_nl = (g_msg == NULL);
-	va_start(vl, format);
-	vsprintf(str, format, vl);
-	g_msg = (char*)realloc(g_msg, (g_msg ? strlen(g_msg) : 0) + strlen(str) + 3);
-	if(add_nl) strcat(g_msg, "\r\n");
-	strcat(g_msg, str);
-#endif
 	va_end(vl);
 }
 
@@ -469,8 +531,9 @@ static void log_message(char* format, ...)
  * @param msg_chunk the message chunk as a null-terminated string
  * @param chunk_size the size of the chunk in bytes
  * @param count the number of chunks in the message
+ * @param set_filename boolean flag: set a filename for the BTIH hash
  */
-static char* repeat_hash(unsigned hash_id, const char* chunk, size_t chunk_size, size_t msg_size)
+static char* repeat_hash(unsigned hash_id, const char* chunk, size_t chunk_size, size_t msg_size, int set_filename)
 {
 	struct rhash_context *ctx;
 	size_t left, size;
@@ -478,14 +541,12 @@ static char* repeat_hash(unsigned hash_id, const char* chunk, size_t chunk_size,
 	assert(rhash_get_hash_length(hash_id) < 130);
 
 	ctx = rhash_init(hash_id);
-#ifdef USE_BTIH_WITH_TEST_FILENAME
-	if((hash_id & RHASH_BTIH) != 0) {
-		unsigned long long total_size = msg_size;
-		rhash_transmit(RMSG_BT_ADD_FILE, ctx, RHASH_STR2UPTR("test.txt"), (rhash_uptr_t)&total_size);
-	}
-#endif
 
-	for(left = msg_size; left > 0; left -= size) {
+	if ((hash_id & RHASH_BTIH) && set_filename) {
+		rhash_torrent_add_file(ctx, "test.txt", (unsigned long long)msg_size);
+	}
+
+	for (left = msg_size; left > 0; left -= size) {
 		size = (left > chunk_size ? chunk_size : left);
 		rhash_update(ctx, (const unsigned char*)chunk, size);
 	}
@@ -505,14 +566,15 @@ static char* repeat_hash(unsigned hash_id, const char* chunk, size_t chunk_size,
  * @param chunk_size the size of the chunk in bytes
  * @param count the number of chunks in the message
  * @param hash the expected hash value
+ * @param set_filename need to set a filename for BTIH hash
  */
-static void assert_hash_long_msg(unsigned hash_id, const char* msg_chunk, size_t chunk_size, size_t msg_size, const char* hash, const char* msg_name)
+static void assert_hash_long_msg(unsigned hash_id, const char* msg_chunk, size_t chunk_size, size_t msg_size, const char* hash, const char* msg_name, int set_filename)
 {
 	char* result;
-	result = repeat_hash(hash_id, msg_chunk, chunk_size, msg_size);
-	if(strcmp(result, hash) != 0) {
+	result = repeat_hash(hash_id, msg_chunk, chunk_size, msg_size, set_filename);
+	if (strcmp(result, hash) != 0) {
 		const char* hash_name = rhash_get_name(hash_id); /* the hash function name */
-		if(msg_name) log_message("failed: %s(%s) = %s, expected %s\n", hash_name, msg_name, result, hash);
+		if (msg_name) log_message("failed: %s(%s) = %s, expected %s\n", hash_name, msg_name, result, hash);
 		else log_message("failed: %s(\"%s\") = %s, expected %s\n", hash_name, msg_chunk, result, hash);
 		g_errors++;
 	}
@@ -523,11 +585,12 @@ static void assert_hash_long_msg(unsigned hash_id, const char* msg_chunk, size_t
  *
  * @param hash_id id of the hash algorithm to use
  * @param msg the message to hash
+ * @param set_filename need to set a filename for BTIH hash
  */
-static char* hash_message(unsigned hash_id, const char* msg)
+static char* hash_message(unsigned hash_id, const char* msg, int set_filename)
 {
 	size_t msg_size = strlen(msg);
-	return repeat_hash(hash_id, msg, msg_size, msg_size);
+	return repeat_hash(hash_id, msg, msg_size, msg_size, set_filename);
 }
 
 /**
@@ -537,11 +600,12 @@ static char* hash_message(unsigned hash_id, const char* msg)
  * @param hash_id id of the algorithm to test
  * @param message the message to hash
  * @param expected_hash the expected hash value
+ * @param set_filename need to set a filename for BTIH hash
  */
-static void assert_hash(unsigned hash_id, const char* msg, const char* expected_hash)
+static void assert_hash(unsigned hash_id, const char* msg, const char* expected_hash, int set_filename)
 {
 	size_t msg_size = strlen(msg);
-	assert_hash_long_msg(hash_id, msg, msg_size, msg_size, expected_hash, NULL);
+	assert_hash_long_msg(hash_id, msg, msg_size, msg_size, expected_hash, NULL, set_filename);
 }
 
 /**
@@ -552,15 +616,16 @@ static void assert_hash(unsigned hash_id, const char* msg, const char* expected_
  * @param ch the character the message is filled with
  * @param msg_size the size of message in bytes
  * @param expected_hash the expected hash value
+ * @param set_filename need to set a filename for BTIH hash
  */
-static void assert_rep_hash(unsigned hash_id, char ch, size_t msg_size, const char* hash)
+static void assert_rep_hash(unsigned hash_id, char ch, size_t msg_size, const char* hash, int set_filename)
 {
-	char ALIGN_ATTR(16) msg_chunk[8192]; /* 8 KiB */
+	char ALIGN_ATTR(64) msg_chunk[8192]; /* 8 KiB */
 	char msg_name[20];
 	memset(msg_chunk, ch, 8192);
-	if(ch >= 32) sprintf(msg_name, "\"%c\"x%d", ch, (int)msg_size);
+	if (ch >= 32) sprintf(msg_name, "\"%c\"x%d", ch, (int)msg_size);
 	else sprintf(msg_name, "\"\\%o\"x%d", (unsigned)(unsigned char)ch, (int)msg_size);
-	assert_hash_long_msg(hash_id, msg_chunk, 8192, msg_size, hash, msg_name);
+	assert_hash_long_msg(hash_id, msg_chunk, 8192, msg_size, hash, msg_name, set_filename);
 }
 
 /*=========================================================================*
@@ -572,11 +637,12 @@ static void assert_rep_hash(unsigned hash_id, char ch, size_t msg_size, const ch
  *
  * @param hash_id id of the algorithm to test
  * @param ptr pointer to array of pairs <message,expected-hash>
+ * @param set_filename need to set a filename for BTIH hash
  */
-static void test_known_strings_pairs(unsigned hash_id, const char** ptr)
+static void test_known_strings_pairs(unsigned hash_id, const char** ptr, int set_filename)
 {
-	for(; ptr[0] && ptr[1]; ptr += 2) {
-		assert_hash(hash_id, ptr[0], ptr[1]);
+	for (; ptr[0] && ptr[1]; ptr += 2) {
+		assert_hash(hash_id, ptr[0], ptr[1], set_filename);
 	}
 }
 
@@ -588,9 +654,10 @@ static void test_known_strings_pairs(unsigned hash_id, const char** ptr)
 static void test_known_strings(unsigned hash_id)
 {
 	int i;
-	for(i = 0; known_strings[i].tests != 0; i++) {
-		if(hash_id == known_strings[i].hash_id) {
-			test_known_strings_pairs(hash_id, known_strings[i].tests);
+	for (i = 0; short_test_vectors[i].tests != 0; i++) {
+		if (hash_id == short_test_vectors[i].hash_id) {
+			int set_filename = (short_test_vectors[i].tests == btih_with_filename_tests);
+			test_known_strings_pairs(hash_id, short_test_vectors[i].tests, set_filename);
 			break;
 		}
 	}
@@ -602,9 +669,9 @@ static void test_known_strings(unsigned hash_id)
 static void test_all_known_strings(void)
 {
 	int i;
-	for(i = 0; known_strings[i].tests != 0; i++) {
-		assert(i < (int)(sizeof(known_strings) / sizeof(known_strings_t)));
-		test_known_strings_pairs(known_strings[i].hash_id, known_strings[i].tests);
+	for (i = 0; short_test_vectors[i].tests != 0; i++) {
+		int set_filename = (short_test_vectors[i].tests == btih_with_filename_tests);
+		test_known_strings_pairs(short_test_vectors[i].hash_id, short_test_vectors[i].tests, set_filename);
 	}
 }
 
@@ -626,6 +693,7 @@ static void test_long_strings(void)
 
 	struct id_to_hash_t tests[] = {
 		{ RHASH_CRC32, "DC25BFBC" }, /* verified with cksfv */
+		{ RHASH_CRC32C, "436FE240" },
 		{ RHASH_MD4, "BBCE80CC6BB65E5C6745E30D4EECA9A4" }, /* checked by md4sum */
 		{ RHASH_MD5, "7707D6AE4E027C70EEA2A935C2296F21" }, /* checked by md5sum */
 		{ RHASH_SHA1, "34AA973CD4C4DAA4F61EEB2BDBAD27316534016F" }, /* checked by sha1sum */
@@ -644,38 +712,44 @@ static void test_long_strings(void)
 		{ RHASH_SHA256, "CDC76E5C9914FB9281A1C7E284D73E67F1809A48A497200E046D39CCC7112CD0" }, /* from NESSIE test vectors */
 		{ RHASH_SHA384, "9D0E1809716474CB086E834E310A4A1CED149E9C00F248527972CEC5704C2A5B07B8B3DC38ECC4EBAE97DDD87F3D8985" }, /* from NESSIE test vectors */
 		{ RHASH_SHA512, "E718483D0CE769644E2E42C7BC15B4638E1F98B13B2044285632A803AFA973EBDE0FF244877EA60A4CB0432CE577C31BEB009C5C2C49AA2E4EADB217AD8CC09B" }, /* from NESSIE test vectors */
-		{ RHASH_SHA3_224, "19F9167BE2A04C43ABD0ED554788101B9C339031ACC8E1468531303F" }, /* verified by refence implementation */
-		{ RHASH_SHA3_256, "FADAE6B49F129BBB812BE8407B7B2894F34AECF6DBD1F9B0F0C7E9853098FC96" }, /* verified by refence implementation */
-		{ RHASH_SHA3_384, "0C8324E1EBC182822C5E2A086CAC07C2FE00E3BCE61D01BA8AD6B71780E2DEC5FB89E5AE90CB593E57BC6258FDD94E17" }, /* verified by refence implementation */
-		{ RHASH_SHA3_512, "5CF53F2E556BE5A624425EDE23D0E8B2C7814B4BA0E4E09CBBF3C2FAC7056F61E048FC341262875EBC58A5183FEA651447124370C1EBF4D6C89BC9A7731063BB" }, /* verified by refence implementation */
+		{ RHASH_SHA3_224, "D69335B93325192E516A912E6D19A15CB51C6ED5C15243E7A7FD653C" },
+		{ RHASH_SHA3_256, "5C8875AE474A3634BA4FD55EC85BFFD661F32ACA75C6D699D0CDCB6C115891C1" },
+		{ RHASH_SHA3_384, "EEE9E24D78C1855337983451DF97C8AD9EEDF256C6334F8E948D252D5E0E76847AA0774DDB90A842190D2C558B4B8340" },
+		{ RHASH_SHA3_512, "3C3A876DA14034AB60627C077BB98F7E120A2A5370212DFFB3385A18D4F38859ED311D0A9D5141CE9CC5C66EE689B266A8AA18ACE8282A0E0DB596C90B0A7B87" },
 		{ RHASH_EDONR256, "56F4B8DC0A41C8EA0A6A42C949883CD5DC25DF8CF4E43AD474FD4492A7A07966" }, /* verified by eBASH SUPERCOP implementation */
 		{ RHASH_EDONR512, "B4A5A255D67869C990FE79B5FCBDA69958794B8003F01FD11E90FEFEC35F22BD84FFA2E248E8B3C1ACD9B7EFAC5BC66616E234A6E938D3526DEE26BD0DE9C562" }, /* verified by eBASH SUPERCOP implementation */
-#ifdef USE_BTIH_WITH_TEST_FILENAME
-		{ RHASH_BTIH, "0DA5C5A6BA0D067C4DE134BB5AD525544DB11A6C" }, /* BTIH with filename="test.txt", verified using uTorrent */
-#else
-		{ RHASH_BTIH, "30CF71DEFC48D497D4C6DCA8FAB203C1E253A53F" }, /* BTIH calculated without a filename, can't be verified by torrent tools */
+		{ RHASH_BTIH, "90AE73EE72A12B5A3A39DCA4C5E24BE1F39B6A1B" } /* BTIH with filename="test.txt", verified using uTorrent */
+#ifdef USE_KECCAK
+		{ RHASH_KECCAK_224, "19F9167BE2A04C43ABD0ED554788101B9C339031ACC8E1468531303F" }, /* verified by reference implementation */
+		{ RHASH_KECCAK_256, "FADAE6B49F129BBB812BE8407B7B2894F34AECF6DBD1F9B0F0C7E9853098FC96" }, /* verified by reference implementation */
+		{ RHASH_KECCAK_384, "0C8324E1EBC182822C5E2A086CAC07C2FE00E3BCE61D01BA8AD6B71780E2DEC5FB89E5AE90CB593E57BC6258FDD94E17" }, /* verified by reference implementation */
+		{ RHASH_KECCAK_512, "5CF53F2E556BE5A624425EDE23D0E8B2C7814B4BA0E4E09CBBF3C2FAC7056F61E048FC341262875EBC58A5183FEA651447124370C1EBF4D6C89BC9A7731063BB" }, /* verified by reference implementation */
 #endif
 	};
 
-	/* test all algorithms on 1,000,000 charaters of 'a' */
-	for(count = 0; count < (sizeof(tests) / sizeof(id_to_hash_t)); count++) {
-		assert_rep_hash(tests[count].hash_id, 'a', 1000000, tests[count].expected_hash);
+	/* test all algorithms on 1,000,000 characters of 'a' */
+	for (count = 0; count < (sizeof(tests) / sizeof(id_to_hash_t)); count++) {
+		int set_filename = (tests[count].hash_id == RHASH_BTIH);
+		assert_rep_hash(tests[count].hash_id, 'a', 1000000, tests[count].expected_hash, set_filename);
 	}
 
-	/* now we verify some specific cases */
-	assert_rep_hash(RHASH_GOST, 0xFF, 64, "13416C4EC74A63C3EC90CB1748FD462C7572C6C6B41844E48CC1184D1E916098");
-	assert_rep_hash(RHASH_GOST_CRYPTOPRO, 0xFF, 64, "58504D26B3677E756BA3F4A9FD2F14B3BA5457066A4AA1D700659B90DCDDD3C6");
+	/* BTIH calculated without a filename. The hash value can't be verified by torrent tools */
+	assert_rep_hash(RHASH_BTIH, 'a', 1000000, "24742F9AE1BD416CF0A6916F2849FE7ABFAC405E", 0);
+
+	/* now we verify some specific cases, which caused problems in many libraries */
+	assert_rep_hash(RHASH_GOST, 0xFF, 64, "13416C4EC74A63C3EC90CB1748FD462C7572C6C6B41844E48CC1184D1E916098", 0);
+	assert_rep_hash(RHASH_GOST_CRYPTOPRO, 0xFF, 64, "58504D26B3677E756BA3F4A9FD2F14B3BA5457066A4AA1D700659B90DCDDD3C6", 0);
 
 	/* these messages verified by eMule LinkCreator (which uses eMule variant of ED2K hash) */
-	assert_rep_hash(RHASH_ED2K, 0, 9728000, "FC21D9AF828F92A8DF64BEAC3357425D");
-	assert_rep_hash(RHASH_ED2K, 0, 9728000 - 1, "AC44B93FC9AFF773AB0005C911F8396F");
-	assert_rep_hash(RHASH_ED2K, 0, 9728000 + 1, "06329E9DBA1373512C06386FE29E3C65"); /* msg with: 9728000 < size <= 9732096 */
-	assert_rep_hash(RHASH_AICH, 0, 9728000, "5D3N4HQHIUMQ7IU7A5QLPLI6RHSWOR7B");
-	assert_rep_hash(RHASH_AICH, 0, 9728000 - 1, "L6SPMD2CM6PRZBGRQ6UFC4HJFFOATRA4");
-	assert_rep_hash(RHASH_AICH, 0, 9728000 + 1, "HL3TFXORIUEPXUWFPY3JLR7SMKGTO4IH");
+	assert_rep_hash(RHASH_ED2K, 0, 9728000, "FC21D9AF828F92A8DF64BEAC3357425D", 0);
+	assert_rep_hash(RHASH_ED2K, 0, 9728000 - 1, "AC44B93FC9AFF773AB0005C911F8396F", 0);
+	assert_rep_hash(RHASH_ED2K, 0, 9728000 + 1, "06329E9DBA1373512C06386FE29E3C65", 0); /* msg with: 9728000 < size <= 9732096 */
+	assert_rep_hash(RHASH_AICH, 0, 9728000, "5D3N4HQHIUMQ7IU7A5QLPLI6RHSWOR7B", 0);
+	assert_rep_hash(RHASH_AICH, 0, 9728000 - 1, "L6SPMD2CM6PRZBGRQ6UFC4HJFFOATRA4", 0);
+	assert_rep_hash(RHASH_AICH, 0, 9728000 + 1, "HL3TFXORIUEPXUWFPY3JLR7SMKGTO4IH", 0);
 #if 0
-	assert_rep_hash(RHASH_ED2K, 0, 9728000 * 5, "3B613901DABA54F6C0671793E28A1205");
-	assert_rep_hash(RHASH_AICH, 0, 9728000 * 5, "EZCO3XF2RJ4FERRDEXGOSSRGL5NA5BBM");
+	assert_rep_hash(RHASH_ED2K, 0, 9728000 * 5, "3B613901DABA54F6C0671793E28A1205", 0);
+	assert_rep_hash(RHASH_AICH, 0, 9728000 * 5, "EZCO3XF2RJ4FERRDEXGOSSRGL5NA5BBM", 0);
 #endif
 }
 
@@ -694,16 +768,15 @@ static void test_results_consistency(void)
 	char res2[70];
 	unsigned i, hash_id;
 
-	for(i = 0, hash_id = 1; (hash_id & RHASH_ALL_HASHES); hash_id <<= 1, i++) {
+	for (i = 0, hash_id = 1; (hash_id & RHASH_ALL_HASHES); hash_id <<= 1, i++) {
 		digest_size = rhash_get_digest_size(hash_id);
 		assert(digest_size < 70);
 
 		ctx = rhash_init(hash_id);
 
-#ifdef USE_BTIH_WITH_TEST_FILENAME
-		if((hash_id & RHASH_BTIH) != 0) {
-			unsigned long long total_size = msg_size;
-			rhash_transmit(RMSG_BT_ADD_FILE, ctx, RHASH_STR2UPTR("test.txt"), (rhash_uptr_t)&total_size);
+#ifndef USE_BTIH_WITH_TEST_FILENAME
+		if ((hash_id & RHASH_BTIH) != 0) {
+			rhash_torrent_add_file(ctx, "test.txt", (unsigned long long)msg_size);
 		}
 #endif
 
@@ -712,7 +785,7 @@ static void test_results_consistency(void)
 		rhash_print(res2, ctx, hash_id, RHPR_RAW);
 		rhash_free(ctx);
 
-		if(memcmp(res1, res2, digest_size) != 0) {
+		if (memcmp(res1, res2, digest_size) != 0) {
 			log_message("failed: inconsistent %s(\"%s\") hash results\n", rhash_get_name(hash_id), msg);
 		}
 	}
@@ -726,27 +799,27 @@ static void test_alignment(void)
 	int i, start, hash_id, alignment_size;
 
 	/* loop by sums */
-	for(i = 0, hash_id = 1; (hash_id & RHASH_ALL_HASHES); hash_id <<= 1, i++) {
+	for (i = 0, hash_id = 1; (hash_id & RHASH_ALL_HASHES); hash_id <<= 1, i++) {
 		char expected_hash[130];
 		assert(rhash_get_digest_size(hash_id) < (int)sizeof(expected_hash));
 
 		alignment_size = (hash_id & (RHASH_TTH | RHASH_TIGER | RHASH_WHIRLPOOL | RHASH_SHA512) ? 8 : 4);
 
 		/* start message with different alignment */
-		for(start = 0; start < alignment_size; start++) {
+		for (start = 0; start < alignment_size; start++) {
 			char message[30];
 			int j, msg_length = 11 + alignment_size;
 
 			/* fill the buffer fifth shifted letter sequence */
-			for(j = 0; j < msg_length; j++) message[start + j] = 'a' + j;
+			for (j = 0; j < msg_length; j++) message[start + j] = 'a' + j;
 			message[start + j] = 0;
 
-			if(start == 0) {
+			if (start == 0) {
 				/* save original hash value */
-				strcpy(expected_hash, hash_message(hash_id, message + start));
+				strcpy(expected_hash, hash_message(hash_id, message + start, 0));
 			} else {
 				/* verify obtained hash value */
-				assert_hash(hash_id, message + start, expected_hash);
+				assert_hash(hash_id, message + start, expected_hash, 0);
 			}
 		}
 	}
@@ -759,7 +832,7 @@ static void test_alignment(void)
 static void test_endianness(void)
 {
 	unsigned tmp = 1;
-	if(*(char*)&tmp == IS_BIG_ENDIAN) {
+	if (*(char*)&tmp == IS_BIG_ENDIAN) {
 		log_message("error: wrong endianness detected at compile time\n");
 		g_errors++;
 	}
@@ -793,15 +866,15 @@ static void assert_magnet(const char* expected,
 	flags &= ~TEST_PATH;
 	size = rhash_print_magnet(out, path, ctx, mask, flags);
 
-	if(expected && strcmp(expected, out) != 0) {
+	if (expected && strcmp(expected, out) != 0) {
 		log_message("error: \"%s\" != \"%s\"\n", expected, out);
 		g_errors++;
 	} else {
 		size_t size2 = strlen(out) + 1;
-		if(size != size2) {
+		if (size != size2) {
 			log_message("error: rhash_print_magnet returns wrong length %d != %d for \"%s\"\n", (int)size, (int)size2, out);
 			g_errors++;
-		} else if(size != (size2 = rhash_print_magnet(NULL, path, ctx, mask, flags))) {
+		} else if (size != (size2 = rhash_print_magnet(NULL, path, ctx, mask, flags))) {
 			log_message("error: rhash_print_magnet(NULL, ...) returns wrong length %d != %d for \"%s\"\n", (int)size2, (int)size, out);
 			g_errors++;
 		}
@@ -821,11 +894,11 @@ static void test_magnet(void)
 
 	assert_magnet("magnet:?xl=1&dn=test.txt&xt=urn:tree:tiger:czquwh3iyxbf5l3bgyugzhassmxu647ip2ike4y", ctx, RHASH_TTH, RHPR_FILESIZE | TEST_PATH);
 	assert_magnet("magnet:?xl=1&xt=urn:md5:0CC175B9C0F1B6A831C399E269772661", ctx, RHASH_MD5, RHPR_FILESIZE | RHPR_UPPERCASE);
-	assert_magnet("xt=urn:ed2k:bde52cb31de33e46245e05fbdbd6fb24&xt=urn:aich:q336in72uwt7zyk5dxolt2xk5i3xmz5y&xt=urn:sha1:q336in72uwt7zyk5dxolt2xk5i3xmz5y&xt=urn:btih:7vai5hicjnmkk6vbge7p6faal74lfror",
+	assert_magnet("xt=urn:ed2k:bde52cb31de33e46245e05fbdbd6fb24&xt=urn:aich:q336in72uwt7zyk5dxolt2xk5i3xmz5y&xt=urn:sha1:q336in72uwt7zyk5dxolt2xk5i3xmz5y&xt=urn:btih:qj6nrgcg7qjs4lth4kocpbggkrb3wtob",
 		ctx, RHASH_ED2K | RHASH_AICH | RHASH_SHA1 | RHASH_BTIH, RHPR_NO_MAGNET);
 
 	/* verify length calculation for all hashes */
-	for(bit = 1; bit < RHASH_ALL_HASHES; bit <<= 1) {
+	for (bit = 1; bit < RHASH_ALL_HASHES; bit <<= 1) {
 		assert_magnet(NULL, ctx, bit, RHPR_FILESIZE | RHPR_NO_MAGNET);
 	}
 
@@ -833,7 +906,7 @@ static void test_magnet(void)
 }
 
 /**
- * Find hash id by its name.
+ * Find a hash function id by its name.
  *
  * @param name hash algorithm name
  * @return algorithm id
@@ -844,22 +917,50 @@ static unsigned find_hash(const char* name)
 	unsigned hash_id;
 	int i;
 
-	if(strlen(name) > (sizeof(buf) - 1)) return 0;
-	for(i = 0; name[i]; i++) buf[i] = toupper(name[i]);
+	if (strlen(name) > (sizeof(buf) - 1)) return 0;
+	for (i = 0; name[i]; i++) buf[i] = toupper(name[i]);
 	buf[i] = 0;
 
-	for(hash_id = 1; (hash_id & RHASH_ALL_HASHES); hash_id <<= 1) {
-		if(strcmp(buf, rhash_get_name(hash_id)) == 0) return hash_id;
+	for (hash_id = 1; (hash_id & RHASH_ALL_HASHES); hash_id <<= 1) {
+		if (strcmp(buf, rhash_get_name(hash_id)) == 0) return hash_id;
 	}
 	return 0;
 }
 
-/* define program entry point */
+/**
+ * Print status of OpenSSL plugin.
+ */
+static void print_openssl_status(void)
+{
+	rhash_uptr_t available = rhash_get_openssl_available_mask();
+	rhash_uptr_t supported = rhash_get_openssl_supported_mask();
+	int has_openssl = rhash_is_openssl_supported();
 
-#ifndef UNDER_CE /* if not Windows CE */
+	printf("OpenSSL %s", (has_openssl ? "supported" : "not supported"));
+	if (has_openssl && available != RHASH_ERROR)
+	{
+		printf(", %s", (available ? "loaded" : "not loaded"));
+		if (available)
+		{
+			unsigned hash_id;
+			printf(":");
+			available &= RHASH_ALL_HASHES;
+			for (hash_id = 1; hash_id <= available; hash_id <<= 1) {
+				if (!!(hash_id & available))
+					printf(" %s", rhash_get_name(hash_id));
+			}
+			supported &= (~available & RHASH_ALL_HASHES);
+			for (hash_id = 1; hash_id <= supported; hash_id <<= 1) {
+				if (!!(hash_id & supported))
+					printf(" -%s", rhash_get_name(hash_id));
+			}
+		}
+	}
+	printf("\n");
+}
 
 /**
- * The application entry point under Linux and Windows.
+ * The program entry point.
  *
  * @param argc number of arguments including the program name
  * @param argv program arguments including the program name
@@ -867,81 +968,38 @@ static unsigned find_hash(const char* name)
  */
 int main(int argc, char *argv[])
 {
-	/* rhash_transmit(RMSG_SET_OPENSSL_MASK, 0, RHASH_ALL_HASHES, 0); */
-
 #ifndef USE_RHASH_DLL
 	rhash_library_init();
 #endif
 
 	test_generic_assumptions();
+	if (argc > 1) {
+		if (strcmp(argv[1], "--speed") == 0) {
+			unsigned hash_id = (argc > 2 ? find_hash(argv[2]) : RHASH_SHA1);
+			if (hash_id == 0) {
+				fprintf(stderr, "error: unknown hash_id: %s\n", argv[2]);
+				return 1;
+			}
+			test_known_strings(hash_id);
 
-	if(argc > 1 && strcmp(argv[1], "--speed") == 0) {
-		unsigned hash_id = (argc > 2 ? find_hash(argv[2]) : RHASH_SHA1);
-		if(hash_id == 0) {
-			fprintf(stderr, "error: unknown hash_id: %s\n", argv[2]);
-			return 1;
+			rhash_run_benchmark(hash_id, 0, stdout);
+		} else if (strcmp(argv[1], "--info") == 0) {
+			printf("%s", compiler_flags);
+			print_openssl_status();
+		} else {
+			printf("Options: [--speed [HASH_NAME]| --info]\n");
 		}
-		test_known_strings(hash_id);
-
-		rhash_run_benchmark(hash_id, 0, stdout);
-	} else if(argc > 1 && strcmp(argv[1], "--flags") == 0) {
-		printf("%s", compiler_flags);
 	} else {
 		test_all_known_strings();
 		test_long_strings();
 		test_alignment();
 		test_results_consistency();
 		test_magnet();
-		if(g_errors == 0) printf("All sums are working properly!\n");
+		if (g_errors == 0) printf("All sums are working properly!\n");
 		fflush(stdout);
 	}
 
-	if(g_errors > 0) printf("%s", compiler_flags);
+	if (g_errors > 0) printf("%s", compiler_flags);
 
 	return (g_errors == 0 ? 0 : 1);
 }
-#else /* UNDER_CE */
-
-#include <windows.h>
-#include <commctrl.h>
-
-/**
- * Convert a single-byte string to a two-byte (wchar_t*).
- *
- * @param str the string to convert
- * @return converted string of wchar_t
- */
-wchar_t *char2wchar(char* str)
-{
-	size_t origsize;
-	wchar_t *wmsg;
-
-	origsize = strlen(str) + 1;
-	wmsg = (wchar_t*)malloc(origsize * 2);
-	mbstowcs(wmsg, str, origsize);
-	return wmsg;
-}
-
-/**
- * The program entry point under Windows CE
- *
- * @param argc number of arguments including program name
- * @param argv program argumants including its name
- */
-int _tmain(int argc, _TCHAR* argv[])
-{
-	wchar_t *wmsg;
-	(void)argc;
-	(void)argv;
-
-	test_known_strings();
-	test_alignment();
-
-	wmsg = char2wchar(g_msg ? g_msg : "Success!\r\nAll sums are working properly.");
-	MessageBox(NULL, wmsg, _T("caption"), MB_OK | MB_ICONEXCLAMATION);
-	free(wmsg);
-	free(g_msg);
-
-	return (g_errors == 0 ? 0 : 1);
-}
-#endif /* UNDER_CE */
